#   pyRCV2: Preferential vote counting
#   Copyright © 2020  Lee Yingtong Li (RunasSudo)
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.

import decimal
import functools
import math

_quantize_exp = 6

def set_dps(dps):
	global _quantize_exp
	_quantize_exp = decimal.Decimal('10') ** -dps

def compatible_types(f):
	@functools.wraps(f)
	def wrapper(self, other):
		if not isinstance(other, Fixed):
			raise ValueError('Attempt to operate on incompatible types')
		return f(self, other)
	return wrapper

class Fixed:
	"""
	Wrapper for Python Decimal (for fixed-point arithmetic)
	"""
	
	ROUND_DOWN = decimal.ROUND_DOWN
	ROUND_HALF_UP = decimal.ROUND_HALF_UP
	ROUND_HALF_EVEN = decimal.ROUND_HALF_EVEN
	ROUND_UP = decimal.ROUND_UP
	
	def __init__(self, val):
		if isinstance(val, Fixed):
			self.impl = val.impl
		else:
			self.impl = decimal.Decimal(val).quantize(_quantize_exp)
	
	def __repr__(self):
		return '<Fixed {}>'.format(str(self.impl))
	def pp(self, dp):
		"""Pretty print to specified number of decimal places"""
		return format(self.impl, '.{}f'.format(dp))
	
	def to_rational(self):
		"""Convert to an instance of Rational"""
		from pyRCV2.numbers import Rational
		return Rational(self.impl)
	
	@compatible_types
	def __add__(self, other):
		return Fixed(self.impl + other.impl)
	@compatible_types
	def __sub__(self, other):
		return Fixed(self.impl - other.impl)
	@compatible_types
	def __mul__(self, other):
		return Fixed(self.impl * other.impl)
	@compatible_types
	def __truediv__(self, other):
		return Fixed(self.impl / other.impl)
	
	@compatible_types
	def __eq__(self, other):
		return self.impl == other.impl
	@compatible_types
	def __ne__(self, other):
		return self.impl != other.impl
	@compatible_types
	def __gt__(self, other):
		return self.impl > other.impl
	@compatible_types
	def __ge__(self, other):
		return self.impl >= other.impl
	@compatible_types
	def __lt__(self, other):
		return self.impl < other.impl
	@compatible_types
	def __le__(self, other):
		return self.impl <= other.impl
	
	def __floor__(self):
		return Fixed(math.floor(self.impl))
	
	def round(self, dps, mode):
		"""Round to the specified number of decimal places, using the ROUND_* mode specified"""
		return Fixed(self.impl.quantize(decimal.Decimal('10') ** -dps, mode))
