#   pyRCV2: Preferential vote counting
#   Copyright © 2020  Lee Yingtong Li (RunasSudo)
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.

from fractions import Fraction
import functools
import math

def compatible_types(f):
	@functools.wraps(f)
	def wrapper(self, other):
		if not isinstance(other, Rational):
			raise ValueError('Attempt to operate on incompatible types')
		return f(self, other)
	return wrapper

class Rational:
	"""
	Wrapper for Python Fraction (rational arithmetic)
	"""
	
	ROUND_DOWN = 0
	ROUND_HALF_UP = 1
	ROUND_HALF_EVEN = 2
	ROUND_UP = 3
	
	def __init__(self, val):
		if isinstance(val, Rational):
			self.impl = val.impl
		else:
			self.impl = Fraction(val)
	
	def __repr__(self):
		return '<Rational {}>'.format(str(self.impl))
	def pp(self, dp):
		"""
		Pretty print to specified number of decimal places
		"""
		# TODO: Work out if there is a better way of doing this
		return format(float(self.impl), '.{}f'.format(dp))
	
	def to_rational(self):
		return self
	
	def to_num(self):
		"""
		Convert to an instance of Num
		"""
		from pyRCV2.numbers import Num
		return Num(self.impl.numerator) / Num(self.impl.denominator)
	
	@compatible_types
	def __add__(self, other):
		return Rational(self.impl + other.impl)
	@compatible_types
	def __sub__(self, other):
		return Rational(self.impl - other.impl)
	@compatible_types
	def __mul__(self, other):
		return Rational(self.impl * other.impl)
	@compatible_types
	def __truediv__(self, other):
		return Rational(self.impl / other.impl)
	
	@compatible_types
	def __eq__(self, other):
		return self.impl == other.impl
	@compatible_types
	def __ne__(self, other):
		return self.impl != other.impl
	@compatible_types
	def __gt__(self, other):
		return self.impl > other.impl
	@compatible_types
	def __ge__(self, other):
		return self.impl >= other.impl
	@compatible_types
	def __lt__(self, other):
		return self.impl < other.impl
	@compatible_types
	def __le__(self, other):
		return self.impl <= other.impl
	
	def __floor__(self):
		return Rational(math.floor(self.impl))
	
	def round(self, dps, mode):
		"""Round to the specified number of decimal places, using the ROUND_* mode specified"""
		factor = Fraction(10) ** dps
		if mode == Rational.ROUND_DOWN:
			return Rational(math.floor(self.impl * factor) / factor)
		elif mode == Rational.ROUND_HALF_UP:
			raise Exception('ROUND_HALF_UP is not implemented in Python Rational context')
		elif mode == Rational.ROUND_HALF_EVEN:
			return Rational(round(self.impl * factor) / factor)
		elif mode == Rational.ROUND_UP:
			return Rational(math.ceil(self.impl * factor) / factor)
		else:
			raise Exception('Invalid rounding mode')
